import logging
import os
import time
from typing import Callable
from warnings import deprecated

import numpy as np
import pyautogui as gui
from PIL import Image

from ._types import Direction, MouseButton
from .box_array import BoxArray
from .constants import LOCATE_AND_CLICK_DELAY, POINTER_SPEED
from .shapes import Box, BoxSpec, Point

logger = logging.getLogger(__name__)


@deprecated("Use `Eelement.locate().click()` instead.")
def locate_and_click(
    reference: Image.Image | str,
    *,
    clicks: int = 1,
    button: MouseButton = "left",
    region: BoxSpec | None = None,
    confidence: float = 0.999,
    grayscale: bool = True,
    offset: int = 0,
    towards: Direction | None = None,
    index: int = 0,
):
    time.sleep(LOCATE_AND_CLICK_DELAY)
    found_region = locate_on_screen(
        reference,
        region=region,
        confidence=confidence,
        grayscale=grayscale,
        limit=index + 1,
    )
    assert found_region is not None, f"Could not locate {reference} on screen."
    assert len(found_region) > index, (
        f"Not enough detections for {reference}: {len(found_region)}"
    )
    if towards is not None:
        target_box = found_region[index].offset(towards, offset)
    else:
        target_box = found_region[index]
    move_and_click(target=target_box, clicks=clicks, button=button)
    time.sleep(LOCATE_AND_CLICK_DELAY)


def move_and_click(
    *, target: BoxSpec | Point, clicks: int = 1, button: MouseButton = "left"
):
    """Move to the center or edge of the region and click.

    The offset is always added to the calculated target point.
    For example, for 'bottom', offset=(0, 5) means 5 pixels below the bottom edge.
    """
    if isinstance(target, Box):
        target = target.center
    elif isinstance(target, Point):
        target = target
    elif isinstance(target, str):
        target = Box.from_spec(target).center
    else:
        raise TypeError(f"Unsupported type for target_box: {type(target)}")

    current = gui.position()
    duration = np.linalg.norm(np.array(target) - np.array(current)) / POINTER_SPEED
    gui.moveTo(*target, float(duration), gui.easeInOutQuad)  # type: ignore
    gui.click(clicks=clicks, button=button)


def locate_on_screen(
    reference: Image.Image | str,
    region: BoxSpec | None = None,
    confidence: float = 0.999,
    grayscale: bool = True,
    limit: int = 1,
    locator: Callable[[Image.Image, Image.Image], BoxArray] | None = None,
) -> BoxArray | None:
    """Locate a region on the screen."""
    if isinstance(reference, str):
        if not os.path.exists(reference):
            raise FileNotFoundError(f"Image file {reference} does not exist.")
        reference = Image.open(reference)
    if locator is None:
        try:
            locations = list(
                gui.locateAllOnScreen(
                    reference,
                    region=Box.from_spec(region).to_tuple() if region else None,
                    grayscale=grayscale,
                    confidence=confidence,
                )
            )
            return BoxArray((Box.from_tuple(loc) for loc in locations[:limit]))
        except gui.ImageNotFoundException:
            return None
        except FileNotFoundError:
            return None
    else:
        screenshot = gui.screenshot(
            region=Box.from_spec(region).to_tuple() if region else None
        )
        logger.info(
            f"Searching in region: {Box.from_spec(region).to_tuple() if region else None}.\nGiven region: {region}"
        )
        detections = locator(reference, screenshot)
        logger.info("total detections: %d", len(detections))
        if len(detections) == 0:
            return None
        else:
            return BoxArray((det.resolve(region) for det in detections[:limit]))



from __future__ import annotations

from typing import Literal

type MouseButton = Literal["left", "right"]
type Direction = (
    Literal[
        "top",
        "left",
        "bottom",
        "right",
        "top-left",
        "top-right",
        "bottom-left",
        "bottom-right",
    ]
    | int
)



from pathlib import Path

import pyautogui as gui

from pyautoguide import ReferenceImageDir, WorkFlow, text

# Initialize reference directory
refs = ReferenceImageDir(Path("examples/saucedemo/references"))

# Create workflow
wf = WorkFlow("SauceDemo")


@wf.navigation(
    text("Username", region="x:2/3 y:(1-2)/3"), text("Products", region="x:1/3 y:1/3")
)
def perform_login(username: str, password: str):
    """Performs the login action to transition from Login to Dashboard."""
    # refs("username").locate().click()
    refs("username").locate().click()
    gui.write(username, interval=0.1)
    gui.press("tab")
    gui.write(password, interval=0.1)
    # text("Swag Labs").locate_and_click(index=1, offset=400, towards="bottom")
    text("Swag Labs").locate(n=2).select(i=1).offset("bottom", 400).click()


@wf.action()
def add_products_to_cart(target: str):
    """Adds products to the cart."""
    if target == "backpack":
        text(
            "Sauce Labs Bike Light", region="x:2/2 y:(2-4)/5", case_sensitive=False
        ).locate().click()
    else:
        text(
            "Sauce Labs Bike Light", region="x:2/2 y:(2-4)/5", case_sensitive=False
        ).locate().click()
    refs("add_to_cart_button").locate(region="x:2/3 y:(2-3)/3").click(clicks=1)


@wf.navigation(
    text("Swag Labs", region="x:2/3 y:1/3"), text("Your Cart", region="x:1/3 y:1/3")
)
def view_cart():
    """Views the cart."""
    # refs("cart_icon").locate().click()
    text("Remove", region="x:2/3 y:3/3").locate().first().find_color(
        color=(226, 35, 26), towards="top-right", region="x:5/5 y:1/4"
    ).click()


@wf.action()
def checkout():
    """Checks out the items in the cart."""
    refs("checkout_button").locate().first().log_screenshot(
        "examples/saucedemo/screenshots/checkout.png"
    ).click()


gui.hotkey("alt", "tab")
wf.expect(
    text("Products", region="x:1/3 y:1/3"),
    username="standard_user",
    password="secret_sauce",
)
wf.invoke("add_products_to_cart", target="backpack")
wf.expect(text("Your Cart", region="x:1/3 y:1/3"))
wf.invoke("checkout")



from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Callable, overload

from ._types import Direction, MouseButton
from .shapes import Box, BoxSpec

class BoxArray(Sequence[Box]):
    """An immutable sequence of Box objects that proxies methods to its contents."""

    def __init__(self, boxes: Iterable[Box] | None = None) -> None: ...
    @overload
    def __getitem__(self, index: int) -> Box: ...
    @overload
    def __getitem__(self, index: slice) -> BoxArray: ...
    def __getitem__(self, index: int | slice) -> Box | BoxArray: ...
    def __len__(self) -> int: ...
    def __add__(self, boxes: object) -> BoxArray: ...

    # `BoxArray` methods
    def first(self) -> Box: ...
    def last(self) -> Box: ...
    def select(self, *, i: int) -> Box: ...
    def pick(self, region: BoxSpec) -> BoxArray: ...
    def filter_by(self, condition: Callable[[Box], bool]) -> BoxArray: ...
    def relative_to(self, direction: Direction, *, of: BoxSpec) -> BoxArray: ...

    # `Box` proxies
    def resolve(self, base: BoxSpec | None = None) -> BoxArray: ...
    def click(self, clicks: int = 1, button: MouseButton = "left") -> BoxArray: ...
    def offset(self, direction: Direction, shift: int = 0) -> BoxArray: ...
    def find_color(
        self,
        color: tuple[int, int, int],
        towards: Direction,
        *,
        region: BoxSpec | None = None,
    ) -> BoxArray: ...



from .references import ImageElement, ReferenceImageDir, TextElement, image, text
from .scene import Scene
from .session import Session
from .shapes import Box, BoxSpec
from .workflow import WorkFlow

__all__ = [
    "Scene",
    "Session",
    "ImageElement",
    "TextElement",
    "BoxSpec",
    "Box",
    "WorkFlow",
    "ReferenceImageDir",
    "text",
    "image",
]



import logging
import time
from random import random
from typing import Callable

import networkx as nx
import pyautogui as gui
from transitions.extensions import GraphMachine

from .references import ReferenceElement
from .utils import get_nx_graph

logger = logging.getLogger(__name__)


class NavigationError(Exception):
    """Custom exception for navigation errors."""

    pass


class WorkFlow:
    """Manages transitions and actions without explicit scenes."""

    def __init__(self, name: str):
        self.name = name
        self.elements: dict[str, ReferenceElement] = {}
        self.navigations: dict[str, Callable] = {}
        self.actions: dict[str, Callable] = {}
        self._sm = GraphMachine()

    def add_element(self, element: ReferenceElement):
        """Add a reference element to the workflow."""
        if element.name not in self.elements:
            self.elements[element.name] = element
            self._sm.add_state(element.name)

    def navigation(self, source: ReferenceElement, to: ReferenceElement):
        """Decorator to define transitions between UI states."""
        self.add_element(source)
        self.add_element(to)

        def decorator[T: Callable](func: T) -> T:
            transition_name = "event_" + func.__name__
            self._sm.add_transition(transition_name, source.name, to.name, prepare=func)
            self.navigations[transition_name] = func
            return func

        return decorator

    def action(self, name: str | None = None):
        """Decorator to define actions that don't change UI state."""

        def decorator[T: Callable](func: T) -> T:
            action_name = name or func.__name__
            self.actions[f"action_{action_name}"] = func
            return func

        return decorator

    def invoke(self, name: str, **kwargs):
        """Execute an action or transition."""
        if (aname := f"action_{name}") in self.actions:
            return self.actions[aname](**kwargs)
        elif (nname := f"event_{name}") in self.navigations:
            return self.navigations[nname](**kwargs)
        raise ValueError(f"Action or navigation '{name}' not found.")

    def get_visible_elements(self) -> list[ReferenceElement]:
        """Return a list of currently visible elements in the workflow."""
        return [
            elem
            for elem in self.elements.values()
            if elem.locate(n=1, error="coerce") is not None
        ]

    def expect(self, elem: ReferenceElement, **kwargs):
        """Navigate to a specific scene."""
        if elem.locate(n=1, error="coerce") is not None:
            return

        graph = get_nx_graph(self._sm)

        visible_elements = self.get_visible_elements()
        all_paths = []
        for present_elem in visible_elements:
            all_paths.extend(
                nx.all_simple_paths(graph, source=present_elem.name, target=elem.name)
            )
        if len(all_paths) == 0:
            raise NavigationError(f"No path found from present screen to {elem}")
        elif len(all_paths) > 1:
            raise NavigationError(f"Multiple paths found from present screen to {elem}")

        path = all_paths[0]
        events: list[str] = [
            graph.get_edge_data(path[i], path[i + 1])[0]["label"]  # type: ignore
            for i in range(len(path) - 1)
        ]
        logger.info(events)

        old_state = self._sm.state
        try:
            self._sm.set_state(path[0])
            for event in events:
                self._sm.dispatch(event, **kwargs)
        except Exception as e:
            self._sm.set_state(old_state)
            raise e

    def wait_for(
        self,
        element: ReferenceElement | list[ReferenceElement],
        timeout: float = 60,
        interval: float = 1,
        keep_busy: bool = True,
    ):
        """Wait until the target scene or reference element is on screen."""
        found = False
        while not found:
            if isinstance(element, ReferenceElement):
                found = element.locate(n=1, error="coerce") is not None
            elif isinstance(element, list):
                for elem in element:
                    assert isinstance(elem, ReferenceElement), (
                        "All elements in the list must be ReferenceElement instances."
                    )
                found = any(elem.locate(n=1) is not None for elem in element)
            else:
                raise TypeError("Target must be a ReferenceElement.")
            start_time = time.time()
            if not found:
                if time.time() - start_time > timeout:
                    raise NavigationError(f"Timeout waiting for {element}")
                w, h = gui.size()
                if keep_busy:
                    gui.moveTo(
                        w * random(),
                        h * random(),
                        duration=2 * interval * random(),
                        tween=gui.easeInOutQuad,  # type: ignore
                    )
                time.sleep(interval)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Literal, overload, override
from warnings import deprecated

import pyautogui as gui
import pyscreeze
from PIL import Image

from ._types import Direction, MouseButton
from .actions import locate_on_screen, move_and_click
from .box_array import BoxArray
from .shapes import Box, BoxSpec
from .utils import get_file


class ElementNotFoundError(Exception):
    """Exception raised when an element is not found on the screen."""

    pass


class ReferenceElement(ABC):
    """Base class for reference elements used to identify scenes."""

    name: str

    @overload
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ) -> BoxArray: ...
    @overload
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ) -> BoxArray | None: ...

    @abstractmethod
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ):
        """Detect the presence of the reference element."""
        raise NotImplementedError("Subclasses must implement this method")

    @deprecated("Use `locate().click()` instead.")
    def locate_and_click(
        self,
        region: BoxSpec | None = None,
        *,
        clicks: int = 1,
        button: MouseButton = "left",
        towards: Direction | None = None,
        offset: int = 0,
        index: int = 0,
    ):
        """Locate the reference element and click on it."""
        regions = self.locate(region=region, n=index + 1)
        assert regions is not None and len(regions) > index, (
            f"Element {self} not found on screen or insufficient detections {len(regions) if regions else 0} < {index + 1}."
        )
        if towards is not None:
            target_box = regions[index].offset(towards, offset)
        else:
            target_box = regions[index]
        move_and_click(target=target_box, clicks=clicks, button=button)


class ImageElement(ReferenceElement):
    """Reference element that identifies a scene by an image."""

    def __init__(
        self,
        path: str | list[str],
        confidence: float = 0.999,
        region: BoxSpec | None = None,
        locator: Callable[[Image.Image, Image.Image], BoxArray] | None = None,
    ):
        self.path = path
        self.confidence = confidence
        self.region = region
        self.locator = locator
        self.name = Path(path).stem if isinstance(path, str) else Path(path[0]).stem

    @overload
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ) -> BoxArray: ...
    @overload
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ) -> BoxArray | None: ...

    @override
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ):
        """Method to detect the presence of the image in the current screen."""
        if isinstance(self.path, str):
            path = [self.path]  # Ensure path is a list for consistency
        else:
            path = self.path

        all_locations: BoxArray = BoxArray()
        for image_path in path:
            try:
                locations = locate_on_screen(
                    image_path,
                    region=region if region else self.region,
                    confidence=self.confidence,
                    locator=self.locator,
                    limit=n - len(all_locations),  # Only get remaining needed locations
                )
                if locations is not None:
                    all_locations += locations

                # If we have enough detections, return them
                if len(all_locations) >= n:
                    return all_locations[:n]
            except (gui.ImageNotFoundException, pyscreeze.ImageNotFoundException):
                continue

        if all_locations:
            return all_locations
        else:
            if error == "coerce":
                return None
            else:
                raise ElementNotFoundError(
                    f"{self} not found on screen in region {region}."
                )

    def __repr__(self) -> str:
        return f"ImageElement: {self.path}"


class ReferenceImageDir:
    def __init__(self, dir_path: Path | str) -> None:
        if isinstance(dir_path, str):
            dir_path = Path(dir_path)
        assert dir_path.is_dir(), f"{dir_path} is not a valid directory."
        self.dir_path = dir_path
        self.images: dict[str, ImageElement] = {}

    def __call__(
        self,
        image_name: str,
        region: BoxSpec | None = None,
        confidence: float = 0.999,
        locator: Callable[[Image.Image, Image.Image], BoxArray] | None = None,
    ) -> ImageElement:
        """Get an ImageElement from the reference directory."""
        if image_name not in self.images:
            image_path = get_file(self.dir_path, name=image_name)
            self.images[image_name] = ImageElement(
                str(image_path), region=region, confidence=confidence, locator=locator
            )
        return self.images[image_name]


def image(
    path: str,
    region: BoxSpec | None = None,
    confidence: float = 0.999,
    locator: Callable[[Image.Image, Image.Image], BoxArray] | None = None,
) -> ImageElement:
    """Create an image reference element."""
    return ImageElement(path, confidence=confidence, region=region, locator=locator)


class TextElement(ReferenceElement):
    """Reference element that identifies a scene by text."""

    def __init__(
        self,
        text: str,
        region: BoxSpec | None = None,
        case_sensitive: bool = False,
        full_text: bool = False,
    ):
        self.text = text
        self.region = region
        self.case_sensitive = case_sensitive
        self.full_text = full_text
        if not case_sensitive:
            self.text = self.text.lower()
        self.name = text

    @overload
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ) -> BoxArray: ...

    @overload
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "coerce",
    ) -> BoxArray | None: ...

    @override
    def locate(
        self,
        region: BoxSpec | None = None,
        n: int = 1,
        error: Literal["raise", "coerce"] = "raise",
    ):
        """Method to detect the presence of the text in the current screen."""
        from .ocr import OCR

        ocr = OCR()
        region = region or self.region
        found_regions = []

        for text, detected_region in ocr.recognize_text(
            gui.screenshot(region=Box.from_spec(region).to_tuple() if region else None)
        ):
            if not self.case_sensitive:
                text = text.lower()
            if self.full_text and text.strip() == self.text.strip():
                found_regions.append(detected_region.resolve(base=region))
            elif not self.full_text and self.text in text:
                found_regions.append(detected_region.resolve(base=region))
            # If we have enough detections, return them
            if len(found_regions) >= n:
                return BoxArray(found_regions[:n])

        if found_regions:
            return BoxArray(found_regions)
        else:
            if error == "coerce":
                return None
            else:
                raise ElementNotFoundError(
                    f"{self} not found on screen in region {region}."
                )

    def __repr__(self) -> str:
        return f"{{TextElement: {self.text}}}"


def text(
    text: str,
    region: BoxSpec | None = None,
    case_sensitive: bool = False,
    full_text: bool = False,
) -> TextElement:
    """Create a text reference element."""
    return TextElement(
        text=text, region=region, case_sensitive=case_sensitive, full_text=full_text
    )
